local addon, RDM = ...

__RDMSAVED = {}

local SILENT=		99
local RELEASE=		90
local TESTING=		20
local WDOGYIELD=	15
local FUNCTIONRUN=	10
local STATUSUPDT=	8
local INFORMATION=	5
local DETAIL=		1

local RDM_INACTIVE=	0
local RDM_SETUP=	1
local RDM_ACTIVE=	2

local RDM_MODE=		RDM_SETUP

local coroDead = false
---------------------
-- Variables
---------------------

local UIElements = {
	cats = {},
	iconbar = {},
	realtime = {},
	cp = {},
	sel = { rows = {} },
	rep = { abilities = {}, headers = {} },
	mmbutton = {}, popupdtl = {},
	sizes = {}
}

RDM.UI = UIElements

local _abilityLookup = {}
local _activeIcons = {}
local _uptimes = {}
local _grpuptime={}
local _grpidx = {}

local _abilityDetails = RDM._abilityDetails

RDM.AD = _abilityDetails

local _checklist = {
	{ name = "-AP/SP I", group=true, list = {
		{ name = "Rift Disturbance" },
		{ name = "Spirit of the Trickster" },
	}},
	{ name = "ARMOR I", group=true, list = {
		{ name = "Lightning Fury" },
		{ name = "Piercing Shot" },
		{ name = "Reaping Touch" },
		{ name = "Splinter Bomb" },
	}},
	{ name = "ARMOR II", group=true, list = {
		{ name = "Crushing Runeshard" },
	}},
	{ name = "ARMOR III", group=true, list = {
		{ name = "Mark of Inevitability" },
	}},	
	{ name = "HASTE I", group=true, list = {
		{ name = "Burning Purpose" },
		{ name = "Call of the Forest" },
	}},
	{ name = "HASTE II", group=true, list = {
		{ name = "Primal Lucidity" },
		{ name = "Surging Flare" },
	}},	
	{ name = "PHYSICAL DMG I", group=true, list = {
		{ name = "Coda of Jeopardy" },
		{ name = "Illuminate" },
		{ name = "Spotter's Call" },
		{ name = "Thorns of Ire" },
	}},
	{ name = "PHYSICAL DMG II", group=true, list = {
		{ name = "Ashen Defense" },
		{ name = "Coda of Cowardice" },
	}},
	{ name = "+STATS", group=true, list = {
		{ name = "Resonance" },
	}},
	{ name = "-DMG OUT", group=true, list = {
		{ name = "Power Drain" },
	}},	
	{ name = "MAGIC DMG I", group=true, list = {
		{ name = "Coda of Distress" },
		{ name = "Crumbling Resistance" },
		{ name = "Spiritual Deficiency" },
	}},
	{ name = "MAGIC DMG II", group=true, list = {
		{ name = "Clinging Spirit" },
		{ name = "Tenebrious Distortion" },
		{ name = "Embers Bomb" },
		{ name = "Sigil of Spirits" },
	}},
	{ name = "MAGIC DMG III", group=true, list = {
		{ name = "Wild Growth" },
	}},	
	{ name = "+CASTING TIME", group=true, list = {
		{ name = "Lingering Dust" },
	}},
	{ name = "+CRITICAL I", group=true, list = {
		{ name = "Lethal Poison" },
		{ name = "Mark of Extermination" },
	}},
	{ name = "+CRITICAL II", group=true, list = {
		{ name = "Call of Blood" },
		{ name = "Dauntless Strike" },
		{ name = "Earthen Barrage" },
	}},
	{ name = "+DAMAGE", group=true, list = {
		{ name = "Call of Savagery" },
		{ name = "Natural Dedication" },
		{ name = "Volcanic Bomb" },
	}},	
	{ name = "-HIT", group=true, list = {
		{ name = "Blood Fever" },
		{ name = "Censure" },
		{ name = "Wrath of the Planes" },
	}},
	{ name = "+RAID DMG", group=false, list = {
		{ name = "Power Core" },
		{ name = "Orchestra of the Planes" },
		{ name = "Lava Field" },
	}},
}

local LANG_LOOKUP = {}

local UINeeded = true
local SVLoaded = false
local RSNeeded = false

UIElements.context = UI.CreateContext(addon.identifier)

RDMonitor_Settings = {}

local default_settings = { trx=200, try=200, selx=300, sely=300, repx=400, repy=400, mmx = 100, mmy = 100, rtx = 600, rty = 200,
	hidden = {}, SAVEDDATA = {}, showtracker = true,
	trk_target = "focus", showdetail = true,
	scale = 1, updatefreq = 0.25, savetime = 60, debuglevel = SILENT, flags = false,
	orient = 'H'
	}

local UPDATEFREQ = default_settings.updatefreq
local SAVETIME = default_settings.savetime

local _playerID = nil

local TIMENOW = 0

local PS_ALIVE		= 1
local PS_DEAD		= 2
local PLAYERSTATE	= PS_ALIVE
local PS_MSG		= { "PS_ALIVE", "PS_DEAD" }

local RS_ALIVE		= 1
local RS_WIPED		= 2
local RS_RECOVER	= 3
local RAIDSTATE		= RS_ALIVE
local RS_MSG		= { "RS_ALIVE", "RS_WIPED", "RS_RECOVER" }

local CB_ACTIVE		= 1
local CB_INACTIVE	= 2
local COMBATSTATE	= CB_INACTIVE
local CB_MSG		= { "CB_ACTIVE", "CB_INACTIVE" }

local CURFIGHTDTLS	= { targetID = 0, targetName = "", ACTIVEBUFFS = {}, st = 0, et = 0, r = "", dx = "" }

local _REPORTSELECTION = false
local _SUMMARYSELECTION = false

local _playerDeathTime = 0

local _chkCPVis = false
local _popupVis = false

local popupCur = ""
local popupPrv = ""

local _watchdogyield = {}
local _debugmessages = {}
RDM.debugmsg = _debugmessages

local debuglevel = default_settings.debuglevel

local function Debug(s,l)
	if (l or INFORMATION) >= debuglevel then
		print(s)
	end
	table.insert(_debugmessages, { tx = os.date("%Y%m%d-%H:%M:%S"), lvl = l, msg = s })
end

local yieldcnt = 0
local function WATCHDOGWAIT(x)
	local rmx = Inspect.System.Watchdog()
	while rmx < 0.03 do
		yieldcnt = yieldcnt+1
		Debug("#"..yieldcnt..":rmx: "..rmx.." ["..x.."]",WDOGYIELD)
		_watchdogyield[yieldcnt] = { tx = os.date("%Y%m%d-%H:%M:%S"), m = x, tr = rmx }
		coroutine.yield()
		rmx = Inspect.System.Watchdog()
	end
end

local function MergeTable(o,n)
	for k,v in pairs(n) do
		if type(v) == "table" then
			if o[k] == nil then
				o[k] = {}			
			end
	 	 	if type(o[k]) == 'table' then
	 			MergeTable(o[k], n[k])
	 	 	end
		else
			if o[k] == nil then
				o[k] = v
			end
		end
	end
end

local _PRVMSG = ""

function RDM.UniqMsg(s)
	if RDMonitor_Settings.flags then
		if s ~= _PRVMSG then
			print(s)
			_PRVMSG = s
		end
	end
end

function RDM.UpdatepopupDtl()
	
	local ix= _grpidx[popupCur]
	local v = _checklist[ix]
	
	popupPrv = popupCur
	popupCur = v.name
	local s = string.format("GROUP: %s", v.name)
	local tn, st
	if COMBATSTATE == CB_ACTIVE then
		tn = TIMENOW
		st = CURFIGHTDTLS.st
	else
		st = 0
		tn = 0
	end
	local fd = tn-st
	Debug(string.format("INCOMBAT: %s. TN=%.2f, TE=%.2f, CS=%.2f, CE=%.2f, fd=%.2f",
		CB_MSG[COMBATSTATE], TIMENOW, tn, st, tn, fd), DETAIL)
	local ut = RDM.GetGroupUptime(v.name)
	local pc = 0
	if fd>0 then
		pc = ut/fd
	end
	s = string.format("%s\nFight Duration:%.2fs\nGroup Uptime:%.2fs (%.2f%%)\n", s, fd, ut, pc*100)

	for k1,v1 in pairs(v.list) do
		local u = 0
		local ud = _uptimes[v1.name]
		if ud ~= nil then
			u = ud.s
			if ud.cs > 0 then
				u = u + tn- ud.cs
			end
		end
		s = string.format("%s\n%s : %.2fs", s, _abilityDetails[v1.name].name, u)
	end
	UIElements.popupdtl:SetText(s)
	UIElements.popupdtl:SetVisible(true)
end

function RDM.ResetIconStatus()
	for k,v in pairs(_abilityLookup) do
		v.f:SetAlpha(0.25)
	end
	for k,v in pairs(_checklist) do
		UIElements.cats[k].activeIcon:SetTexture("Rift", "Data/\\UI\\texture\\global\\placeholder_icon.dds")
		UIElements.cats[k].activeIcon:SetAlpha(0.25)
		UIElements.cats[k].frameIcon:SetBackgroundColor(0,0,0,1)
		UIElements.cats[k].frame:SetWidth(0)
	end
end

function RDM.FindValidTarget()
	local d = Inspect.Unit.Detail(RDMonitor_Settings.trk_target)
	if d ~= nil then
		if d.availability == "full" then
			if d.relation == "hostile" then
				if d.health > 0 then
					RDM.StartFight(d.id, d.name)
				end
			end
		end
	end
end

function RDM.StartFight(id, name)
	CURFIGHTDTLS = { targetID = id, targetName = name, st = TIMENOW, et = TIMENOW, r = "", ACTIVEBUFFS = {}, dx = os.date("%Y%m%d-%H:%M:%S") }
	_uptimes = {}
	_grpuptime = {}
	for k,v in pairs(_checklist) do
		_grpuptime[v.name] = {pct=0, l={}}
	end

end

function RDM.EndFight(r)
	CURFIGHTDTLS.et = TIMENOW
	CURFIGHTDTLS.r = r

	for k,v in pairs(_uptimes) do
		local _cs = _uptimes[k].cs
		if _cs > 0 then
			local b =_abilityLookup[k]
			table.insert(_uptimes[k].l, {s = _cs, e = TIMENOW, d = (TIMENOW-_cs) })
			table.insert(_grpuptime[b.s].l, {s = _cs, e = TIMENOW})
			_uptimes[k].s = _uptimes[k].s + (TIMENOW-_cs)
		end
	end

	if (CURFIGHTDTLS.et - CURFIGHTDTLS.st) > SAVETIME then
		Debug(string.format("Saving fight vs. %s dur %.2f [%s]", CURFIGHTDTLS.targetName, (CURFIGHTDTLS.et - CURFIGHTDTLS.st), CURFIGHTDTLS.r), RELEASE)
		local cr = {
			rn = CURFIGHTDTLS.r,
			tx = CURFIGHTDTLS.dx,
			tn = CURFIGHTDTLS.targetName,
			fd = (CURFIGHTDTLS.et-CURFIGHTDTLS.st),
			cs = CURFIGHTDTLS.st,
			ce = CURFIGHTDTLS.et,
			utg = {}, utd = {}, dtl = { ut = {}, gu = {} }
		}
		for k,v in pairs(_checklist) do
			cr.utg[v.name] = RDM.GetGroupUptime(v.name)
		end
		for k,v in pairs(_uptimes) do
			cr.utd[k] = v.s
		end
		cr.dtl.ut = _uptimes
		cr.dtl.gu = _grpuptime
		table.insert(RDMonitor_Settings.SAVEDDATA, cr)
		if _REPORTSELECTION == true then
			RDM.UpdateReportSelection(UIElements.sel.page)
		end
	else
		Debug(string.format("Not saving fight. Duration %.2f < %d", (CURFIGHTDTLS.et - CURFIGHTDTLS.st), SAVETIME), RELEASE)
	end
	RDM.ResetIconStatus()
	CURFIGHTDTLS = { targetID = 0, targetName = "", ACTIVEBUFFS = {}, st = TIMENOW, et = 0, r = "", dx = "" }
end

function RDM.HandleTarget()
	local id = CURFIGHTDTLS.targetID
	local d = Inspect.Unit.Detail(id)

	if d ~= nil then
		if d.availability == "full" then
			if d.health > 0 then
				local _curBuffs = {}
				local bl = Inspect.Buff.List(id) or {}
				for k,v in pairs(Inspect.Buff.Detail(id, bl)) do
					local bn = LANG_LOOKUP[v.name]
					if bn ~= nil then
						if _abilityDetails[bn].id == nil then
							_curBuffs[bn] = true
						else
							if _abilityDetails[bn].id == v.abilityNew then
								_curBuffs[bn] = true
							end
						end
					end
				end
				
				bl = Inspect.Buff.List("player") or {}
				for k,v in pairs(Inspect.Buff.Detail("player", bl)) do
					local bn = LANG_LOOKUP[v.name]
					if bn ~= nil then
						_curBuffs[bn] = true
					end
				end
				
				local _gain = {}
				local _loss = {}
			
				for k,v in pairs(CURFIGHTDTLS.ACTIVEBUFFS) do
					if _curBuffs[k] == nil then
						Debug("Lost buff: "..k, TESTING)
						_loss[k] = true
					end
				end
				for k,v in pairs(_curBuffs) do
					if CURFIGHTDTLS.ACTIVEBUFFS[k] == nil then
						Debug("Gained buff:"..k, TESTING)
						_gain[k] = true
					end
				end
				for k,_ in pairs(_loss) do
					CURFIGHTDTLS.ACTIVEBUFFS[k] = nil
					if _uptimes[k] == nil then
						Debug(string.format("NOTICE: LOSS %s, but no matching found. Ignoring.", k), RELEASE)
					else
						local _cs = _uptimes[k].cs
						if _cs > 0 then
							Debug(string.format("CU:Adding buff=%s, start = %.2f, end = %.2f, dur = %.2f", k, _cs, TIMENOW, (TIMENOW-_cs)), TESTING)
							table.insert(_uptimes[k].l, {s = _cs, e = TIMENOW, d = (TIMENOW-_cs) })
							local b =_abilityLookup[k]
							if _activeIcons[b.ix] == k then
								b.f:SetAlpha(0.50)
								UIElements.cats[b.ix].activeIcon:SetAlpha(0.25)
							end
							table.insert(_grpuptime[b.s].l, {s = _cs, e = TIMENOW})
							_uptimes[k].s = _uptimes[k].s + (TIMENOW-_cs)
							_uptimes[k].cs = 0
						else
							Debug(string.format("NOTICE: buff=%s, start = 0. IGNORING.", k), RELEASE)
						end
					end
				end
				for k,_ in pairs(_gain) do
					CURFIGHTDTLS.ACTIVEBUFFS[k] = true
					if _uptimes[k] == nil then
						_uptimes[k] = { s=0, cs = 0, l = {} }
						Debug(string.format("NOTICE: creating _uptimes entry for %s", k), TESTING)
					end
					_uptimes[k].cs = TIMENOW
					local b =_abilityLookup[k]
					b.f:SetAlpha(1)
					UIElements.cats[b.ix].activeIcon:SetAlpha(1)
					UIElements.cats[b.ix].activeIcon:SetTexture("Rift", _abilityDetails[k].i)
					_activeIcons[b.ix] = k
				end
				local fd = TIMENOW - CURFIGHTDTLS.st

				for k,v in pairs(_checklist) do
					if RDMonitor_Settings.hidden[v.name] == nil then
						local uptime = RDM.GetGroupUptime(v.name)
						local pct
						if fd > 0 then
							pct = uptime/fd
							if pct > 1 then
								Debug("NOTICE: >100% uptime on: "..v.name, TESTING)
								pct = 1
							end
						else
							pct = 0
						end
						local lpct= _grpuptime[v.name].pct
						if pct > lpct then
							UIElements.cats[k].frame:SetBackgroundColor(0,0.5,0)
							UIElements.cats[k].frameIcon:SetBackgroundColor(0,0.5,0,1)
						elseif pct < lpct then
							UIElements.cats[k].frame:SetBackgroundColor(0.5,0,0)
							UIElements.cats[k].frameIcon:SetBackgroundColor(0.5,0,0,1)
						elseif pct==0 then
							UIElements.cats[k].frameIcon:SetBackgroundColor(0,0,0,1)
						end
						_grpuptime[v.name].pct = pct
						UIElements.cats[k].frame:SetWidth(pct * UIElements.sizes.frW)
					end
				end
				if _popupVis then
					RDM.UpdatepopupDtl()
				end
			else
				Debug("Target is a corpse...", STATUSUPDT)
			end
		else
			Debug("Target only partially available.", STATUSUPDT)
		end
	else
		RDM.EndFight("NOTARGETCHAIN")
	end
end

function RDM.CheckRaid()
	local s = string.format("[%s][%s][%s]", (PS_MSG[PLAYERSTATE] or "??"), (CB_MSG[COMBATSTATE] or "??"), (RS_MSG[RAIDSTATE] or "??"))
	if PLAYERSTATE == PS_ALIVE and COMBATSTATE == CB_INACTIVE then
		RDM.UniqMsg(string.format("%s : DO NOTHING", s))
		if CURFIGHTDTLS.targetID ~= 0 then
			RDM.EndFight("RANAWAY")
		end
	elseif PLAYERSTATE == PS_ALIVE and COMBATSTATE == CB_ACTIVE then
		if CURFIGHTDTLS.targetID == 0 then
			RDM.UniqMsg(string.format("%s : LOOKING FOR TARGET", s))
			RDM.FindValidTarget()
		end
		if CURFIGHTDTLS.targetID ~= 0 then
			RDM.UniqMsg(string.format("%s : TARGET = %s (%s)", s, CURFIGHTDTLS.targetName, CURFIGHTDTLS.targetID))
			RDM.HandleTarget()
		end
	elseif PLAYERSTATE == PS_DEAD and RAIDSTATE == RS_ALIVE then
		if CURFIGHTDTLS.targetID ~= 0 then
			RDM.UniqMsg(string.format("%s : TARGET = %s (%s)", s, CURFIGHTDTLS.targetName, CURFIGHTDTLS.targetID))
			RDM.HandleTarget()
		end	
		if (os.time() - _playerDeathTime) > 2 then
			local d = Inspect.Unit.Detail("player")
			if d ~= nil then
				if d.availability == "full" then
					if d.health > 0 then
						Debug(string.format("PLAYER ALIVE! %s",tostring(d.health)), STATUSUPDT)
						PLAYERSTATE = PS_ALIVE
					end
				end
			end
		end
		local newRAIDSTATE = RS_WIPED
		for x = 1, 20 do
			d = Inspect.Unit.Detail(string.format("group%02d", x))
			if d ~= nil then
				if d.availability == "full" then
					if (d.health > 0) and (d.relation == "friendly") then
						newRAIDSTATE = RS_ALIVE
						break
					end
				end
			end
		end
		RAIDSTATE = newRAIDSTATE
		if RAIDSTATE == RS_WIPED then
			Debug("RS : ALIVE - WIPE", STATUSUPDT)
			RDM.EndFight("RAIDWIPE")
		end
	elseif PLAYERSTATE == PS_DEAD and RAIDSTATE == RS_WIPED then
		for x = 1, 20 do
			d = Inspect.Unit.Detail(string.format("group%02d", x))
			if d ~= nil then
				if d.availability == "full" then
					if (d.health > 0) and (d.relation == "friendly") then
						Debug("RS : WIPE - RECOVER", STATUSUPDT)
						RAIDSTATE = RS_RECOVER
						break
					end
				end
			end
		end
	else
		RDM.UniqMsg(string.format("%s : NO ACTION DEFINED", s))
	end
end

function RDM.UpdateReportSelection(p)
	local _sets = {}
	for k,v in ipairs(RDMonitor_Settings.SAVEDDATA) do
		table.insert(_sets, {ix=k, tx=(v.tx or "00000000-0000"), tn=(v.tn or "UNKNOWN TARGET")})
	end
	table.sort(_sets, function(a,b) return a.tx > b.tx end)

	local _l = p*16
	local _f = _l - 15
	
	if (p > 1) and (_f > #_sets) then
		p = p-1
		_l = p*16
		_f = _l - 15
	end
	UIElements.sel.page = p
	if p == 1 then
		UIElements.sel.prev.active = false
		UIElements.sel.prev:SetAlpha(0.25)
	else
		UIElements.sel.prev.active = true
		UIElements.sel.prev:SetAlpha(1)
	end

	if (p*16) < #_sets then
		UIElements.sel.next.active = true
		UIElements.sel.next:SetAlpha(1)
	else
		UIElements.sel.next.active = false
		UIElements.sel.next:SetAlpha(0.25)
	end

	local mx

	if _l <= #_sets then
		mx = 16
	else
		mx = #_sets % 16
	end

	for x=1, mx do
		local _setix = _f + x -1
		UIElements.sel.rows[x].name:SetText(string.format("%d. %s", _setix, _sets[_setix].tn))
		UIElements.sel.rows[x].date:SetText(string.sub(_sets[_setix].tx, 5))
		UIElements.sel.rows[x].frame:SetVisible(true)
		UIElements.sel.rows[x].data = _sets[_setix]
	end
	for x=(mx+1), 16 do
		UIElements.sel.rows[x].frame:SetVisible(false)
	end
end

function RDM.ShowSummary(ix)
	local d = RDMonitor_Settings.SAVEDDATA[ix]
	local fd = d.fd or 0
	local w = UIElements.rep.abw
	UIElements.rep.target:SetText(string.format("TARGET: %s [%s]  ", (d.tn or "UNKNOWN TARGET"), (d.rn or "UNKNOWN")))
	UIElements.rep.time:SetText(string.format("Fight start: %s, Duration = %.2fs", (d.tx or "00000000-0000"), fd))

	for k, v in ipairs(_checklist) do
		local ut = d.utg[v.name] or 0
		if fd > 0 then
			if fd > ut then
				UIElements.rep.headers[k].uptime:SetText(string.format("Uptime: %.2fs [%.2f%%]", ut, 100*(ut/fd)))
			else
				UIElements.rep.headers[k].uptime:SetText(string.format("Uptime: %.2fs [100%%]", ut))
			end
		else
			UIElements.rep.headers[k].uptime:SetText(string.format("Uptime: %.2fs", ut))
		end
	end

	for k,v in pairs(_abilityDetails) do
		if d.utd[k] ~= nil then
			local ut = d.utd[k]
			local pc
			if ut > fd then
				pc = 1
			else
				pc = ut/fd
			end
			UIElements.rep.abilities[k].prog:SetWidth(pc * w)
			UIElements.rep.abilities[k].text:SetText(string.format("%s : %.2fs [%.2f%%]", v.name, ut, (100*pc)))
		else
			UIElements.rep.abilities[k].prog:SetWidth(0)
			UIElements.rep.abilities[k].text:SetText(string.format("%s : 0.00s [0.00%%]", v.name))
		end
	end
	if _SUMMARYSELECTION == false then
		_SUMMARYSELECTION = true
		UIElements.rep.header:SetVisible(_SUMMARYSELECTION)
	end
end

function RDM.Scale(f)
	local s14 = 14 * f
	local s16 = 16 * f
	local s18 = 18 * f
	UIElements.realtime.title:SetFontSize(s14)
	UIElements.sizes.h14 = UIElements.realtime.title:GetHeight()
	UIElements.realtime.title:SetFontSize(s16)
	UIElements.sizes.h16 = UIElements.realtime.title:GetHeight()
	UIElements.realtime.title:SetFontSize(s18)
	UIElements.sizes.h18 = UIElements.realtime.title:GetHeight()

	UIElements.popupdtl:SetFontSize(s16)
	
	UIElements.realtime.title:SetFontSize(s18)

	UIElements.sizes.frW = nil

	for k,v in ipairs(_checklist) do
		UIElements.cats[k].frame:SetHeight(UIElements.sizes.h16)
		UIElements.cats[k].bkgrnd:SetHeight(UIElements.sizes.h16)
		UIElements.cats[k].label:SetFontSize(s16)
		UIElements.cats[k].frameIcon:SetHeight(UIElements.sizes.h14+4)
		UIElements.cats[k].frameIcon:SetWidth(UIElements.sizes.h14+4)
		UIElements.cats[k].activeIcon:SetHeight(UIElements.sizes.h14)
		UIElements.cats[k].activeIcon:SetWidth(UIElements.sizes.h14)
		if UIElements.sizes.frW == nil then
			UIElements.cats[k].label:SetText("+CASTING COSTXXX")
			UIElements.sizes.frW = UIElements.cats[k].label:GetWidth()
			UIElements.cats[k].label:SetText(v.name)
		end
		UIElements.cats[k].bkgrnd:SetWidth(UIElements.sizes.frW)
		for k1,v1 in pairs(UIElements.cats[k].icons) do
			v1:SetHeight(UIElements.sizes.h16)
			v1:SetWidth(UIElements.sizes.h16)
		end
	end

	UIElements.iconbar.options:SetHeight(UIElements.sizes.h14+4)
	UIElements.iconbar.options:SetWidth(UIElements.sizes.h14+4)
	
	UIElements.realtime.frame:SetWidth(UIElements.sizes.frW+(UIElements.sizes.h16*6)+12)
	UIElements.realtime.title:SetWidth(UIElements.realtime.frame:GetWidth())
	UIElements.popupdtl:SetWidth(UIElements.sizes.frW+(UIElements.sizes.h16*6)+12)
	UIElements.popupdtl:SetPoint("BOTTOMRIGHT", UI.Native.TooltipAnchor, "BOTTOMRIGHT", 0,0)

	UIElements.sizes.h18m = UIElements.sizes.h18*1.75
	
	UIElements.realtime.close:SetHeight(UIElements.sizes.h18)
	UIElements.realtime.close:SetWidth(UIElements.sizes.h18)
	
	UIElements.cp.zoom:SetHeight(UIElements.sizes.h18m)
	UIElements.cp.zoom:SetWidth(UIElements.sizes.h18m)
	UIElements.cp.rotate:SetHeight(UIElements.sizes.h18m)
	UIElements.cp.rotate:SetWidth(UIElements.sizes.h18m)
	UIElements.cp.scope:SetHeight(UIElements.sizes.h18m)
	UIElements.cp.scope:SetWidth(UIElements.sizes.h18m)
	UIElements.cp.scopetxt:SetFontSize(24*f)
	UIElements.cp.list:SetHeight(UIElements.sizes.h18m)
	UIElements.cp.list:SetWidth(UIElements.sizes.h18m)
	UIElements.cp.update:SetHeight(UIElements.sizes.h18m)
	UIElements.cp.update:SetWidth(UIElements.sizes.h18m)
	UIElements.cp.save:SetHeight(UIElements.sizes.h18m)
	UIElements.cp.save:SetWidth(UIElements.sizes.h18m)
	UIElements.cp.report:SetHeight(UIElements.sizes.h18m)
	UIElements.cp.report:SetWidth(UIElements.sizes.h18m)
	UIElements.cp.stop:SetHeight(UIElements.sizes.h18m)
	UIElements.cp.stop:SetWidth(UIElements.sizes.h18m)
	
	RDM.OrderFrames()
end

function RDM.GetGroupUptime(g)
	local nt = {}
	local t = _grpuptime[g].l
	local ut = 0
	table.sort(t, function(a,b) return a.s < b.s end)
	local cs=0
	local ce=0
	for k,v in ipairs(t) do
		if k==1 then
			cs = v.s
			ce = v.e
		else
			if v.s <= ce then
				if v.e > ce then
					ce = v.e
				end
			elseif v.s > ce then
				table.insert(nt, {s=cs, e=ce})
				Debug(string.format("[%s] NEW PERIOD: %.2f - %.2f = %.2f", g,cs, ce, ce-cs), DETAIL)
				ut = ut + (ce-cs)
				cs = v.s
				ce = v.e
			end
		end
	end
	if cs > 0 then
		table.insert(nt, {s=cs, e=ce})
		Debug(string.format("[%s] NEW PERIOD: %.2f - %.2f = %.2f", g,cs, ce, ce-cs), DETAIL)
		ut = ut + (ce-cs)
	end
	local ix = _grpidx[g]
	local ns = TIMENOW
	for k,v in pairs(_checklist[ix].list) do
		if _uptimes[v.name] ~= nil then
			local acs = _uptimes[v.name].cs
			if acs > 0 and acs < ns then
				ns = acs
			end
		end
	end
	if ns < TIMENOW then
		if ns < ce then
			Debug(string.format("[%s] ACT OVERLAP: %.2f/%.2f - %.2f = %.2f", g,ce,ns, TIMENOW, TIMENOW-ce), DETAIL)
			ut = ut + (TIMENOW - ce)
		else
			Debug(string.format("[%s] ACT NEW PERIOD: %.2f - %.2f = %.2f", g, ns, TIMENOW, TIMENOW-ns), DETAIL)
			ut = ut + (TIMENOW - ns)
		end
	end
	Debug(string.format("Uptime of %s is %f", g, ut), DETAIL)
	return ut
end

function RDM.ResizeCP(in_x, in_y)
	UIElements.cp.container:ClearPoint("TOPLEFT")
	UIElements.cp.container:ClearPoint("TOPRIGHT")
	UIElements.cp.container:ClearPoint("BOTTOMLEFT")
	UIElements.cp.zoom:ClearPoint("BOTTOMLEFT")
	UIElements.cp.zoom:ClearPoint("TOPLEFT")
	UIElements.cp.zoom:ClearPoint("TOPRIGHT")

	if RDMonitor_Settings.orient == 'H' then
		UIElements.cp.container:SetWidth((UIElements.sizes.h18m*8)+4)
		UIElements.cp.container:SetHeight(UIElements.sizes.h18m + UIElements.sizes.h14+6)
		if in_y < 50 then
			UIElements.cp.pos = "BOTTOM"
			UIElements.cp.container:SetPoint("TOPLEFT", UIElements.iconbar.options, "TOPLEFT", 0,0)
			UIElements.cp.zoom:SetPoint("BOTTOMLEFT", UIElements.cp.container, "BOTTOMLEFT", 2,-2)
		else
			UIElements.cp.pos = "TOP"
			UIElements.cp.container:SetPoint("BOTTOMLEFT", UIElements.iconbar.options, "BOTTOMLEFT", 0,0)
			UIElements.cp.zoom:SetPoint("TOPLEFT", UIElements.cp.container, "TOPLEFT", 2, 2)
		end
		UIElements.cp.rotate:SetPoint("TOPLEFT", UIElements.cp.zoom, "TOPRIGHT", 0,0)
		UIElements.cp.list:SetPoint("TOPLEFT", UIElements.cp.rotate, "TOPRIGHT", 0,0)
		UIElements.cp.update:SetPoint("TOPLEFT", UIElements.cp.list, "TOPRIGHT", 0,0)
		UIElements.cp.save:SetPoint("TOPLEFT", UIElements.cp.update, "TOPRIGHT", 0,0)
		UIElements.cp.report:SetPoint("TOPLEFT", UIElements.cp.save, "TOPRIGHT", 0,0)
		UIElements.cp.stop:SetPoint("TOPLEFT", UIElements.cp.report, "TOPRIGHT", 0,0)
		UIElements.cp.scope:SetPoint("TOPLEFT", UIElements.cp.stop, "TOPRIGHT", 0,0)
	else
		UIElements.cp.container:SetHeight((UIElements.sizes.h18m*8)+4)
		UIElements.cp.container:SetWidth(UIElements.sizes.h18m + UIElements.sizes.h14+6)
		if in_x < 50 then
			UIElements.cp.pos = "RIGHT"
			UIElements.cp.container:SetPoint("TOPLEFT", UIElements.iconbar.options, "TOPLEFT", 0,0)
			UIElements.cp.zoom:SetPoint("TOPRIGHT", UIElements.cp.container, "TOPRIGHT", -2, -2)
		else
			UIElements.cp.pos = "LEFT"
			UIElements.cp.container:SetPoint("TOPRIGHT", UIElements.iconbar.options, "TOPRIGHT", 0,0)
			UIElements.cp.zoom:SetPoint("TOPLEFT", UIElements.cp.container, "TOPLEFT", 2, 2)
		end
		UIElements.cp.rotate:SetPoint("TOPLEFT", UIElements.cp.zoom, "BOTTOMLEFT", 0,0)
		UIElements.cp.list:SetPoint("TOPLEFT", UIElements.cp.rotate, "BOTTOMLEFT", 0,0)
		UIElements.cp.update:SetPoint("TOPLEFT", UIElements.cp.list, "BOTTOMLEFT", 0,0)
		UIElements.cp.save:SetPoint("TOPLEFT", UIElements.cp.update, "BOTTOMLEFT", 0,0)
		UIElements.cp.report:SetPoint("TOPLEFT", UIElements.cp.save, "BOTTOMLEFT", 0,0)
		UIElements.cp.stop:SetPoint("TOPLEFT", UIElements.cp.report, "BOTTOMLEFT", 0,0)
		UIElements.cp.scope:SetPoint("TOPLEFT", UIElements.cp.stop, "BOTTOMLEFT", 0,0)
	end
end

function RDM.OrderFrames()
	local prev = nil
	local cnt = 1
	for k,v in pairs(_checklist) do
		if RDMonitor_Settings.hidden[v.name] == nil then
			cnt = cnt+1
			if prev == nil then
				UIElements.cats[k].frame:SetPoint("TOPLEFT", UIElements.realtime.frame, "TOPLEFT", (UIElements.sizes.h16*6)+12,UIElements.sizes.h18)
				if RDMonitor_Settings.orient == 'H' then
					UIElements.cats[k].frameIcon:SetPoint("TOPLEFT", UIElements.iconbar.options, "TOPRIGHT", 2,0)
				else
					UIElements.cats[k].frameIcon:SetPoint("TOPLEFT", UIElements.iconbar.options, "BOTTOMLEFT", 0,2)
				end
			else
				UIElements.cats[k].frame:SetPoint("TOPLEFT", UIElements.cats[prev].frame, "BOTTOMLEFT", 0,2)
				if RDMonitor_Settings.orient == 'H' then
					UIElements.cats[k].frameIcon:SetPoint("TOPLEFT", UIElements.cats[prev].frameIcon, "TOPRIGHT", 2,0)
				else
					UIElements.cats[k].frameIcon:SetPoint("TOPLEFT", UIElements.cats[prev].frameIcon, "BOTTOMLEFT", 0,2)
				end
			end
			UIElements.cats[k].frame:SetVisible(true)
			UIElements.cats[k].frameIcon:SetVisible(true)
			prev = k
		else
			UIElements.cats[k].frame:SetVisible(false)
			UIElements.cats[k].frameIcon:SetVisible(false)
		end
	end
	if RDMonitor_Settings.orient == 'H' then
		UIElements.iconbar.frame:SetHeight(UIElements.sizes.h14+8)
		UIElements.iconbar.frame:SetWidth((cnt*(UIElements.sizes.h14+6))+2)
	else
		UIElements.iconbar.frame:SetHeight((cnt*(UIElements.sizes.h14+6))+2)
		UIElements.iconbar.frame:SetWidth(UIElements.sizes.h14+8)
	end
	UIElements.realtime.frame:SetHeight((cnt*UIElements.sizes.h16)+(cnt*2)+2+UIElements.sizes.h18)
end

function RDM.SVLoaded()
	if RDMonitor_Settings == nil then RDMonitor_Settings = {} end
	if RDMonitor_Items == nil then RDMonitor_Items = {} end
	MergeTable(RDMonitor_Settings, default_settings)

	for k,v in pairs(RDMonitor_Settings) do
		if default_settings[k] == nil then
			RDMonitor_Settings[k] = nil
		end
	end

	if not MINIMAPDOCKER then
		UIElements.mmbutton.button:SetPoint("TOPLEFT", UIParent, "TOPLEFT", RDMonitor_Settings.mmx, RDMonitor_Settings.mmy)
	end
	RDM.Scale(RDMonitor_Settings.scale)
	RDM.ResizeCP(RDMonitor_Settings.trx, RDMonitor_Settings.try)
	UIElements.iconbar.frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", RDMonitor_Settings.trx, RDMonitor_Settings.try)

	UIElements.sel.header:SetPoint("TOPLEFT", UIParent, "TOPLEFT", RDMonitor_Settings.selx, RDMonitor_Settings.sely)
	UIElements.rep.header:SetPoint("TOPLEFT", UIParent, "TOPLEFT", RDMonitor_Settings.repx, RDMonitor_Settings.repy)
	UIElements.realtime.frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", RDMonitor_Settings.rtx, RDMonitor_Settings.rty)
	UIElements.realtime.frame:SetVisible(RDMonitor_Settings.showdetail)
	if RDMonitor_Settings.debuglevel ~= default_settings.debuglevel then
		Debug(string.format("Changing debug level from %d to %d", debuglevel, RDMonitor_Settings.debuglevel), SILENT)
		debuglevel = RDMonitor_Settings.debuglevel
	end
	UPDATEFREQ = RDMonitor_Settings.updatefreq
	SAVETIME = RDMonitor_Settings.savetime
	local ssn = #RDMonitor_Settings.SAVEDDATA
	if ssn > 256 then
		Debug(string.format("Have %d saved sets. Need to clear out oldest %d", ssn, (ssn-16)), TESTING)
		for x=(ssn-256), 1, -1 do
			Debug(string.format("Remove: RDMonitor_Settings.SAVEDDATA[%d]", x), TESTING)
			table.remove(RDMonitor_Settings.SAVEDDATA, x)
		end
	else
		Debug(string.format("Have %d saved sets. Nothing to do.", ssn), TESTING)
	end
	
	if RDMonitor_Settings.trk_target == "player.target" then
		UIElements.cp.scopetxt:SetText("T")
	elseif RDMonitor_Settings.trk_target == "player.target.target" then
		UIElements.cp.scopetxt:SetText("T.T")
	else
		UIElements.cp.scopetxt:SetText("F")
	end
end

function RDM.UpdateHandler()
	local lastupdate = 0
	local lastitemchk = 0
	
	local _status
	local _itmDtl

	while true do
		if RDM_MODE == RDM_ACTIVE then
			if _chkCPVis == true then
				_chkCPVis = RDM.chkFrameVis(UIElements.cp.container)
			end
			if RSNeeded then
				RDM.Scale(RDMonitor_Settings.scale)
				RSNeeded = false
			end

			TIMENOW = Inspect.Time.Real()
			
			if TIMENOW - lastupdate > UPDATEFREQ then
				RDM.CheckRaid()
				lastupdate = TIMENOW
			end

		elseif RDM_MODE == RDM_SETUP then
			if UINeeded then
				RDM.CreateUI()
				UINeeded = false
			elseif SVLoaded then
				RDM.SVLoaded()
				SVLoaded = false
				if RDMonitor_Settings.showtracker == true then
					RDM_MODE = RDM_ACTIVE
				else
					RDM_MODE = RDM_INACTIVE
				end
			end
		end
		coroutine.yield()
	end
end

RDM.coUpdateHandler = coroutine.create(RDM.UpdateHandler)

function RDM.Event_System_Update_Begin(h)
	if coroDead then return end
	local ok, errormsg = coroutine.resume(RDM.coUpdateHandler)
	if not ok then
		coroDead = true
		error("Error in coroutine: "..errormsg)
	end
end

function RDM.chkFrameVis(f)
	local md = Inspect.Mouse()
	local l = f:GetLeft()
	local r = f:GetRight()
	local t = f:GetTop()
	local b = f:GetBottom()

	if md.x < l or md.x > r or md.y < t or md.y > b then
		f:SetVisible(false)
		return false
	else
		return true
	end
end

function RDM.CreateUI()

	UIElements.mmbutton.button = UI.CreateFrame("Texture", "UIElements.mmbutton.button", UIElements.context)
	UIElements.mmbutton.button:SetVisible(true)
	UIElements.mmbutton.button:SetLayer(10)
	UIElements.mmbutton.button:SetTexture(addon.identifier, "textures/mm_button.png")
	UIElements.mmbutton.button:SetHeight(36)
	UIElements.mmbutton.button:SetWidth(36)

	if MINIMAPDOCKER then
		MINIMAPDOCKER.Register(addon.identifier, UIElements.mmbutton.button)
	else
		UIElements.mmbutton.button:EventAttach(Event.UI.Input.Mouse.Left.Down, function(self, h)
			self.MouseDown = true
			local mouseData = Inspect.Mouse()
			self.sx = mouseData.x - UIElements.mmbutton.button:GetLeft()
			self.sy = mouseData.y - UIElements.mmbutton.button:GetTop()
		end, "Event.UI.Input.Mouse.Left.Down")

		UIElements.mmbutton.button:EventAttach(Event.UI.Input.Mouse.Cursor.Move, function(self, h)
			if self.MouseDown then
				local nx, ny
				local mouseData = Inspect.Mouse()
				nx = mouseData.x - self.sx
				ny = mouseData.y - self.sy
				UIElements.mmbutton.button:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx,ny)
			end
		end, "Event.UI.Input.Mouse.Cursor.Move")

		UIElements.mmbutton.button:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self, h)
			if self.MouseDown then
				self.MouseDown = false
			end
			RDMonitor_Settings.mmx = UIElements.mmbutton.button:GetLeft()
			RDMonitor_Settings.mmy = UIElements.mmbutton.button:GetTop()
		end, "Event.UI.Input.Mouse.Left.Up")
	end
	
	UIElements.mmbutton.button:EventAttach(Event.UI.Input.Mouse.Right.Click, function(self, h)
		RDMonitor_Settings.showtracker = not RDMonitor_Settings.showtracker
		UIElements.iconbar.frame:SetVisible(RDMonitor_Settings.showtracker)
		if RDMonitor_Settings.showtracker == true then
			RDM_MODE = RDM_ACTIVE
		else
			RDM_MODE = RDM_INACTIVE
			UIElements.iconbar.frame:SetVisible(false)
			UIElements.cp.container:SetVisible(false)
			UIElements.popupdtl:SetVisible(false)
			_REPORTSELECTION = false
			UIElements.sel.header:SetVisible(_REPORTSELECTION)
			_SUMMARYSELECTION = false
			UIElements.rep.header:SetVisible(_SUMMARYSELECTION)
			_chkCPVis = false
		end
	end, "Event.UI.Input.Mouse.Right.Click")
	
	WATCHDOGWAIT("CreateUI:1")
	
	UIElements.realtime.frame = UI.CreateFrame("Frame", "UIElements.realtime.frame", UIElements.context)
	UIElements.realtime.frame:SetLayer(1)
	
	UIElements.realtime.title = UI.CreateFrame("Text", "UIElements.realtime.title", UIElements.realtime.frame)
	UIElements.realtime.title:SetText("RDT RealTime view...")
	UIElements.realtime.title:SetLayer(2)
	UIElements.realtime.title:SetPoint("TOPCENTER", UIElements.realtime.frame, "TOPCENTER", 0,0)
	UIElements.realtime.title:SetBackgroundColor(0,0,0.75)

	UIElements.realtime.title:EventAttach(Event.UI.Input.Mouse.Left.Down, function(self, h)
		self.MouseDown = true
		local mouseData = Inspect.Mouse()
		self.sx = mouseData.x - UIElements.realtime.frame:GetLeft()
		self.sy = mouseData.y - UIElements.realtime.frame:GetTop()
	end, "Event.UI.Input.Mouse.Left.Down")

	UIElements.realtime.title:EventAttach(Event.UI.Input.Mouse.Cursor.Move, function(self, h)
		if self.MouseDown then
			local nx, ny
			local mouseData = Inspect.Mouse()
			nx = mouseData.x - self.sx
			ny = mouseData.y - self.sy
			UIElements.realtime.frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx,ny)
		end
	end, "Event.UI.Input.Mouse.Cursor.Move")

	UIElements.realtime.title:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self, h)
		if self.MouseDown then
			self.MouseDown = false
		end
		RDMonitor_Settings.rtx = UIElements.realtime.frame:GetLeft()
		RDMonitor_Settings.rty = UIElements.realtime.frame:GetTop()
	end, "Event.UI.Input.Mouse.Left.Up")
	
	WATCHDOGWAIT("CreateUI:3")
	
	UIElements.realtime.close = UI.CreateFrame("Texture", "UIElements.realtime.close", UIElements.realtime.frame)
	UIElements.realtime.close:SetTexture(addon.identifier, "textures/cancel.png")
	UIElements.realtime.close:SetPoint("TOPRIGHT", UIElements.realtime.title, "TOPRIGHT", 0,0)
	UIElements.realtime.close:SetLayer(3)
	
	UIElements.realtime.close:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		RDMonitor_Settings.showdetail = false
		UIElements.realtime.frame:SetVisible(false)
	end, "Event.UI.Input.Mouse.Left.Click")
	
	WATCHDOGWAIT("CreateUI:6")

	UIElements.popupdtl = UI.CreateFrame("Text", "UIElements.popupdtl", UIElements.context)
	UIElements.popupdtl:SetVisible(false)
	UIElements.popupdtl:SetLayer(100)
	UIElements.popupdtl:SetBackgroundColor(0.25,0,0.25)

	UIElements.iconbar.frame = UI.CreateFrame("Frame", "UIElements.iconbar.frame", UIElements.context)
	UIElements.iconbar.frame:SetLayer(1)
	UIElements.iconbar.frame:SetBackgroundColor(0.25,0.25,0.25,0.25)
	
	UIElements.iconbar.iconcontainer = UI.CreateFrame("Frame", "UIElements.iconbar.iconcontainer", UIElements.iconbar.frame)
	UIElements.iconbar.iconcontainer:SetPoint("TOPLEFT", UIElements.iconbar.frame, "TOPLEFT", 2,2)
	UIElements.iconbar.iconcontainer:SetLayer(2)
	
	UIElements.iconbar.options = UI.CreateFrame("Texture", "UIElements.iconbar.options", UIElements.iconbar.frame)
	UIElements.iconbar.options:SetTexture("Rift", "Data/\\UI\\texture\\global\\placeholder_icon.dds")
	UIElements.iconbar.options:SetPoint("TOPLEFT", UIElements.iconbar.iconcontainer, "TOPLEFT", 0,0)
	UIElements.iconbar.options:SetLayer(5)

	UIElements.iconbar.options:EventAttach(Event.UI.Input.Mouse.Left.Down, function(self, h)
		self.MouseDown = true
		local mouseData = Inspect.Mouse()
		self.sx = mouseData.x - UIElements.iconbar.frame:GetLeft()
		self.sy = mouseData.y - UIElements.iconbar.frame:GetTop()
	end, "Event.UI.Input.Mouse.Left.Down")

	UIElements.iconbar.options:EventAttach(Event.UI.Input.Mouse.Cursor.Move, function(self, h)
		if self.MouseDown then		
			local nx, ny
			local mouseData = Inspect.Mouse()
			nx = mouseData.x - self.sx
			ny = mouseData.y - self.sy
			UIElements.iconbar.frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx,ny)

			if RDMonitor_Settings.orient == 'H' then
				if mouseData.y < 50 and UIElements.cp.pos ~= "BOTTOM" then
					RDM.ResizeCP(mouseData.x, mouseData.y)
				elseif mouseData.y >= 50 and UIElements.cp.pos ~= "TOP" then
					RDM.ResizeCP(mouseData.x, mouseData.y)
				end
			else
				if mouseData.x < 50 and UIElements.cp.pos ~= "RIGHT" then
					RDM.ResizeCP(mouseData.x, mouseData.y)
				elseif mouseData.x >= 50 and UIElements.cp.pos ~= "LEFT" then
					RDM.ResizeCP(mouseData.x, mouseData.y)
				end
			end
		end
	end, "Event.UI.Input.Mouse.Cursor.Move")

	UIElements.iconbar.options:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self, h)
		if self.MouseDown then
			self.MouseDown = false
		end
		RDMonitor_Settings.trx = UIElements.iconbar.frame:GetLeft()
		RDMonitor_Settings.try = UIElements.iconbar.frame:GetTop()
	end, "Event.UI.Input.Mouse.Left.Up")

	UIElements.iconbar.options:EventAttach(Event.UI.Input.Mouse.Right.Click, function(self, h)
		RDMonitor_Settings.hidden = {}
		RDM.OrderFrames()
	end, "Event.UI.Input.Mouse.Right.Click")
	
	UIElements.iconbar.options:EventAttach(Event.UI.Input.Mouse.Cursor.In, function(self, h)
		UIElements.cp.container:SetVisible(true)
		_chkCPVis = true
	end, "Event.UI.Input.Mouse.Cursor.In")

	-- build ability data sets
	for k,v in ipairs(_checklist) do
		_grpidx[v.name] = k
		WATCHDOGWAIT("CreateUI:7:A")
		_grpuptime[v.name] = { pct=0, l={} }
		local f = {}
		f.frame = UI.CreateFrame("Frame", "f.frame", UIElements.realtime.frame)
		f.frame:SetVisible(false)
		f.frame:SetWidth(0)
		f.frame:SetBackgroundColor(0,0.5,0)
		f.frame:SetLayer(2)
		
		f.frameIcon = UI.CreateFrame("Frame", "f.frameIcon", UIElements.iconbar.frame)
		f.frameIcon:SetBackgroundColor(0,0,0,1)
		f.frameIcon:SetLayer(5)

		f.activeIcon = UI.CreateFrame("Texture", "f.activeIcon", f.frameIcon)
		f.activeIcon:SetAlpha(0.25)
		f.activeIcon:SetLayer(2)
		f.activeIcon:SetTexture("Rift", "Data/\\UI\\texture\\global\\placeholder_icon.dds")
		f.activeIcon:SetPoint("TOPLEFT", f.frameIcon, "TOPLEFT", 2,2)
		f.activeIcon:EventAttach(Event.UI.Input.Mouse.Cursor.In, function(self, h)
			popupCur = v.name
			RDM.UpdatepopupDtl()
			_popupVis = true
		end, "Event.UI.Input.Mouse.Cursor.In")
		f.activeIcon:EventAttach(Event.UI.Input.Mouse.Cursor.Out, function(self, h)
			if popupCur == v.name then
				UIElements.popupdtl:SetVisible(false)
				_popupVis = false
			end
		end, "Event.UI.Input.Mouse.Cursor.Out")
		WATCHDOGWAIT("CreateUI:7:B")
		f.bkgrnd = UI.CreateFrame("Frame", "f.bkgrnd", f.frame)
		f.bkgrnd:SetPoint("TOPLEFT", f.frame, "TOPLEFT", 0,0)
		f.bkgrnd:SetBackgroundColor(0,0,0,0.5)
		f.label = UI.CreateFrame("Text", "f.label", f.frame)
		f.label:SetPoint("TOPLEFT", f.frame, "TOPLEFT", 0,0)
		f.label:SetText(v.name)
		f.label:SetLayer(3)
		function f.bkgrnd.Event.RightClick()
			RDMonitor_Settings.hidden[v.name] = true
			RDM.OrderFrames()
		end
		f.icons = {}
	
		for k1, v1 in ipairs(v.list) do
			local i = UI.CreateFrame("Texture", "f.icon", f.frame)
		
			if _abilityLookup[v1.name] == nil then
				_abilityLookup[v1.name] = { f=i, s=v.name, ix=k }
			end
			i:SetAlpha(0.25)
			if _abilityDetails[v1.name] ~= nil then
				i:SetTexture("Rift", _abilityDetails[v1.name].i)
			else
				i:SetTexture("Rift", "Data/\\UI\\texture\\global\\placeholder_icon.dds")
			end
			i:SetLayer(3)
			if k1 == 1 then
				i:SetPoint("TOPRIGHT", f.label, "TOPLEFT", -2,0)
			else
				i:SetPoint("TOPRIGHT", f.icons[k1-1], "TOPLEFT", -2,0)
			end
			f.icons[k1] = i
		end	

		UIElements.cats[k] = f
	end
	WATCHDOGWAIT("CreateUI:8")

	UIElements.cp.container = UI.CreateFrame("Frame", "UIElements.cp.container", UIElements.context)
	UIElements.cp.container:SetVisible(false)
	UIElements.cp.container:SetBackgroundColor(0,0,0.75)
	UIElements.cp.container:SetLayer(0)

	UIElements.cp.zoom = UI.CreateFrame("Texture", "UIElements.cp.zoom", UIElements.cp.container)
	UIElements.cp.zoom:SetTexture(addon.identifier, "textures/magnifier.png")
	UIElements.cp.zoom:SetLayer(2)
	UIElements.cp.zoom:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		RDMonitor_Settings.scale = RDMonitor_Settings.scale-0.05
		RSNeeded = true
	end, "Event.UI.Input.Mouse.Wheel.Back")
	UIElements.cp.zoom:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		RDMonitor_Settings.scale = RDMonitor_Settings.scale+0.05
		RSNeeded = true
	end, "Event.UI.Input.Mouse.Wheel.Forward")
	UIElements.cp.zoom:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		RDMonitor_Settings.scale = 1
		RSNeeded = true
	end, "Event.UI.Input.Mouse.Left.Click")
	WATCHDOGWAIT("CreateUI:9")
	
	UIElements.cp.rotate = UI.CreateFrame("Texture", "UIElements.cp.rotate", UIElements.cp.container)
	UIElements.cp.rotate:SetTexture(addon.identifier, "textures/page_white_swoosh.png")
	UIElements.cp.rotate:SetLayer(2)
	UIElements.cp.rotate:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		RDM.Command_Slash_Register(0, "orientation")
	end, "Event.UI.Input.Mouse.Left.Click")

	UIElements.cp.list = UI.CreateFrame("Texture", "UIElements.cp.list", UIElements.cp.container)
	UIElements.cp.list:SetTexture(addon.identifier, "textures/application_side_list.png")
	UIElements.cp.list:SetLayer(2)
	UIElements.cp.list:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		RDM.Command_Slash_Register(0, "showdetail")
	end, "Event.UI.Input.Mouse.Left.Click")

	UIElements.cp.update = UI.CreateFrame("Texture", "UIElements.cp.update", UIElements.cp.container)
	UIElements.cp.update:SetTexture(addon.identifier, "textures/time_go.png")
	UIElements.cp.update:SetLayer(2)
	UIElements.cp.update:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		if UPDATEFREQ > 0.05 then
			RDMonitor_Settings.updatefreq = RDMonitor_Settings.updatefreq-0.05
			Debug(string.format("Update frequency changed from %.2fs to %.2fs", UPDATEFREQ, RDMonitor_Settings.updatefreq), SILENT)
			UPDATEFREQ = RDMonitor_Settings.updatefreq
		end
	end, "Event.UI.Input.Mouse.Wheel.Back")
	UIElements.cp.update:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		RDMonitor_Settings.updatefreq = RDMonitor_Settings.updatefreq+0.05
		Debug(string.format("Update frequency changed from %.2fs to %.2fs", UPDATEFREQ, RDMonitor_Settings.updatefreq), SILENT)
		UPDATEFREQ = RDMonitor_Settings.updatefreq
	end, "Event.UI.Input.Mouse.Wheel.Forward")
	UIElements.cp.update:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		Debug(string.format("Update frequency changed from %.2fs to %.2fs", UPDATEFREQ, default_settings.updatefreq), SILENT)
		RDMonitor_Settings.updatefreq = default_settings.updatefreq
		UPDATEFREQ = RDMonitor_Settings.updatefreq
	end, "Event.UI.Input.Mouse.Left.Click")
	WATCHDOGWAIT("CreateUI:10")
	UIElements.cp.save = UI.CreateFrame("Texture", "UIElements.cp.save", UIElements.cp.container)
	UIElements.cp.save:SetTexture(addon.identifier, "textures/time_add.png")
	UIElements.cp.save:SetLayer(2)
	UIElements.cp.save:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		if SAVETIME >= 10 then
			RDMonitor_Settings.savetime = RDMonitor_Settings.savetime-5
			Debug(string.format("Minimum fight duration to save changed from %ds to %ds", SAVETIME, RDMonitor_Settings.savetime), SILENT)
			SAVETIME = RDMonitor_Settings.savetime
		end
	end, "Event.UI.Input.Mouse.Wheel.Back")
	UIElements.cp.save:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		RDMonitor_Settings.savetime = RDMonitor_Settings.savetime+5
		Debug(string.format("Minimum fight duration to save changed from %ds to %ds", SAVETIME, RDMonitor_Settings.savetime), SILENT)
		SAVETIME = RDMonitor_Settings.savetime
	end, "Event.UI.Input.Mouse.Wheel.Forward")
	UIElements.cp.save:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		RDMonitor_Settings.savetime = default_settings.savetime
		Debug(string.format("Minimum fight duration to save changed from %ds to %ds", SAVETIME, RDMonitor_Settings.savetime), SILENT)
		SAVETIME = RDMonitor_Settings.savetime
	end, "Event.UI.Input.Mouse.Left.Click")

	UIElements.cp.report = UI.CreateFrame("Texture", "UIElements.cp.report", UIElements.cp.container)
	UIElements.cp.report:SetTexture(addon.identifier, "textures/report.png")
	UIElements.cp.report:SetLayer(2)
	UIElements.cp.report:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		_REPORTSELECTION = not _REPORTSELECTION
		if _REPORTSELECTION == true then
			RDM.UpdateReportSelection(1)
		end
		UIElements.sel.header:SetVisible(_REPORTSELECTION)
	end, "Event.UI.Input.Mouse.Left.Click")
	WATCHDOGWAIT("CreateUI:11")
	UIElements.cp.scope = UI.CreateFrame("Texture", "UIElements.cp.scope", UIElements.cp.container)
	UIElements.cp.scope:SetTexture(addon.identifier, "textures/scope.png")
	UIElements.cp.scope:SetLayer(2)
	UIElements.cp.scope:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		RDM.Command_Slash_Register(0, "mode")
	end, "Event.UI.Input.Mouse.Left.Click")
	
	UIElements.cp.scopetxt = UI.CreateFrame("Text", "UIElements.cp.scopetxt", UIElements.cp.container)
	UIElements.cp.scopetxt:SetLayer(3)
	UIElements.cp.scopetxt:SetFontColor(1,1,1)
	UIElements.cp.scopetxt:SetPoint("CENTER", UIElements.cp.scope, "CENTER", 0,0)
	
	WATCHDOGWAIT("CreateUI:12")

	UIElements.cp.stop = UI.CreateFrame("Texture", "UIElements.cp.stop", UIElements.cp.container)
	UIElements.cp.stop:SetTexture(addon.identifier, "textures/clock_stop.png")
	UIElements.cp.stop:SetLayer(2)
	UIElements.cp.stop:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		if CURFIGHTDTLS.targetID ~= 0 then
			RDM.EndFight("STOPBUTTON")
		end
	end, "Event.UI.Input.Mouse.Left.Click")

	local sw = 400
	local y18 = 0
	local y16 = 0
	local y14 = 0
	local y12 = 0

	UIElements.sel.header = UI.CreateFrame("Frame", "UIElements.sel.header", UIElements.context)
	UIElements.sel.header:SetVisible(false)
	UIElements.sel.header:SetWidth(sw)
	UIElements.sel.header:SetLayer(1)
	UIElements.sel.header:SetBackgroundColor(0,0,0.5)

	UIElements.sel.header:EventAttach(Event.UI.Input.Mouse.Left.Down, function(self, h)
		self.MouseDown = true
		local mouseData = Inspect.Mouse()
		self.sx = mouseData.x - UIElements.sel.header:GetLeft()
		self.sy = mouseData.y - UIElements.sel.header:GetTop()
	end, "Event.UI.Input.Mouse.Left.Down")

	UIElements.sel.header:EventAttach(Event.UI.Input.Mouse.Cursor.Move, function(self, h)
		if self.MouseDown then
			local nx, ny
			local mouseData = Inspect.Mouse()
			nx = mouseData.x - self.sx
			ny = mouseData.y - self.sy
			UIElements.sel.header:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx,ny)
		end
	end, "Event.UI.Input.Mouse.Cursor.Move")

	UIElements.sel.header:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self, h)
		if self.MouseDown then
			self.MouseDown = false
		end
		RDMonitor_Settings.selx = UIElements.sel.header:GetLeft()
		RDMonitor_Settings.sely = UIElements.sel.header:GetTop()
	end, "Event.UI.Input.Mouse.Left.Up")

	WATCHDOGWAIT("CreateUI:13")

	UIElements.sel.title = UI.CreateFrame("Text", "UIElements.sel.title", UIElements.sel.header)
	UIElements.sel.title:SetText("Debuff Tracker Saved Data")
	UIElements.sel.title:SetLayer(2)
	UIElements.sel.title:SetFontSize(12)
	y12 = UIElements.sel.title:GetHeight()
	UIElements.sel.title:SetFontSize(14)
	y14 = UIElements.sel.title:GetHeight()
	UIElements.sel.title:SetFontSize(16)
	y16 = UIElements.sel.title:GetHeight()
	UIElements.sel.title:SetFontSize(18)
	y18 = UIElements.sel.title:GetHeight()
	UIElements.sel.header:SetHeight(y18)
	UIElements.sel.title:SetPoint("TOPLEFT", UIElements.sel.header, "TOPLEFT")

	UIElements.sel.close = UI.CreateFrame("Texture", "UIElements.sel.close", UIElements.sel.header)
	UIElements.sel.close:SetTexture(addon.identifier, "textures/cancel.png")
	UIElements.sel.close:SetPoint("TOPRIGHT", UIElements.sel.header, "TOPRIGHT", 0,0)
	UIElements.sel.close:SetHeight(y18)
	UIElements.sel.close:SetWidth(y18)
	UIElements.sel.close:SetLayer(2)
	 UIElements.sel.close:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		_REPORTSELECTION = false
		UIElements.sel.header:SetVisible(_REPORTSELECTION)
	end, "Event.UI.Input.Mouse.Left.Click")
	WATCHDOGWAIT("CreateUI:13a")

	UIElements.sel.next = UI.CreateFrame("Texture", "UIElements.sel.next", UIElements.sel.header)
	UIElements.sel.next:SetTexture(addon.identifier, "textures/resultset_next.png")
	UIElements.sel.next:SetPoint("TOPRIGHT", UIElements.sel.close, "TOPLEFT", -20,0)
	UIElements.sel.next:SetHeight(y18)
	UIElements.sel.next:SetWidth(y18)
	UIElements.sel.next:SetLayer(2)
	UIElements.sel.next:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		if UIElements.sel.next.active == true then
			RDM.UpdateReportSelection(UIElements.sel.page+1)
		end
	end, "Event.UI.Input.Mouse.Left.Click")

	UIElements.sel.prev = UI.CreateFrame("Texture", "UIElements.sel.prev", UIElements.sel.header)
	UIElements.sel.prev:SetTexture(addon.identifier, "textures/resultset_previous.png")
	UIElements.sel.prev:SetPoint("TOPRIGHT", UIElements.sel.next, "TOPLEFT", -2,0)
	UIElements.sel.prev:SetHeight(y18)
	UIElements.sel.prev:SetWidth(y18)
	UIElements.sel.prev:SetLayer(2)
	UIElements.sel.prev:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		if UIElements.sel.prev.active == true then
			RDM.UpdateReportSelection(UIElements.sel.page-1)
		end
	end, "Event.UI.Input.Mouse.Left.Click")

	UIElements.sel.bkgrnd = UI.CreateFrame("Frame", "UIElements.sel.bkgrnd", UIElements.sel.header)
	UIElements.sel.bkgrnd:SetPoint("TOPLEFT", UIElements.sel.header, "BOTTOMLEFT", 0,0)
	UIElements.sel.bkgrnd:SetBackgroundColor(0.2, 0.2, 0.2, 0.75)
	UIElements.sel.bkgrnd:SetWidth(sw)
	UIElements.sel.bkgrnd:SetHeight((y16*16)+4)
	UIElements.sel.bkgrnd:SetLayer(1)

	for x=1, 16 do
		WATCHDOGWAIT("CreateUI:14:"..x)
		local f = {}
		f.frame = UI.CreateFrame("Frame", "f.frame", UIElements.sel.header)
		f.frame:SetWidth(sw-4)
		f.frame:SetHeight(y16)
		f.frame:SetLayer(2)
		f.frame:EventAttach(Event.UI.Input.Mouse.Cursor.In, function(self, h)
			f.frame:SetBackgroundColor(0,0.5,0,1)
		end, "Event.UI.Input.Mouse.Cursor.In")
		f.frame:EventAttach(Event.UI.Input.Mouse.Cursor.Out, function(self, h)
			f.frame:SetBackgroundColor(0,0,0,0.25)
		end, "Event.UI.Input.Mouse.Cursor.Out")
		f.frame:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
			RDM.ShowSummary(f.data.ix)
		end, "Event.UI.Input.Mouse.Left.Click")
		f.frame:EventAttach(Event.UI.Input.Mouse.Right.Click, function(self, h)
			table.remove(RDMonitor_Settings.SAVEDDATA, f.data.ix)
			RDM.UpdateReportSelection(UIElements.sel.page)
		end, "Event.UI.Input.Mouse.Right.Click")
	
		if x == 1 then
			f.frame:SetPoint("TOPLEFT", UIElements.sel.bkgrnd, "TOPLEFT", 2,2)
		else
			f.frame:SetPoint("TOPLEFT", UIElements.sel.rows[x-1].frame, "BOTTOMLEFT", 0,0)
		end
	
		f.name = UI.CreateFrame("Text", "f.name", f.frame)
		f.name:SetFontSize(14)
		--f.name:SetText("Some quite Unfeasibly long name of a boss")
		f.name:SetText("")
		f.name:SetLayer(3)
		f.name:SetPoint("CENTERLEFT", f.frame, "CENTERLEFT", 0,0)
	
		f.date = UI.CreateFrame("Text", "f.date", f.frame)
		f.date:SetFontSize(16)
		--f.date:SetText("8888-8888")
		f.date:SetText("")
		f.date:SetLayer(3)
		f.date:SetPoint("TOPRIGHT", f.frame, "TOPRIGHT", 0,0)
	
		f.frame:SetBackgroundColor(0,0,0,0.25)
	
		UIElements.sel.rows[x] = f
	end

	WATCHDOGWAIT("CreateUI:15")

	local sw = 600

	UIElements.rep.header = UI.CreateFrame("Frame", "UIElements.rep.header", UIElements.context)
	UIElements.rep.header:SetVisible(false)
	UIElements.rep.header:SetWidth(sw)
	UIElements.rep.header:SetLayer(1)
	UIElements.rep.header:SetBackgroundColor(0,0,0.5)

	UIElements.rep.header:EventAttach(Event.UI.Input.Mouse.Left.Down, function(self, h)
		self.MouseDown = true
		local mouseData = Inspect.Mouse()
		self.sx = mouseData.x - UIElements.rep.header:GetLeft()
		self.sy = mouseData.y - UIElements.rep.header:GetTop()
	end, "Event.UI.Input.Mouse.Left.Down")

	UIElements.rep.header:EventAttach(Event.UI.Input.Mouse.Cursor.Move, function(self, h)
		if self.MouseDown then
			local nx, ny
			local mouseData = Inspect.Mouse()
			nx = mouseData.x - self.sx
			ny = mouseData.y - self.sy
			UIElements.rep.header:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx,ny)
		end
	end, "Event.UI.Input.Mouse.Cursor.Move")

	UIElements.rep.header:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self, h)
		if self.MouseDown then
			self.MouseDown = false
		end
		RDMonitor_Settings.repx = UIElements.rep.header:GetLeft()
		RDMonitor_Settings.repy = UIElements.rep.header:GetTop()
	end, "Event.UI.Input.Mouse.Left.Up")

	WATCHDOGWAIT("CreateUI:16")

	UIElements.rep.title = UI.CreateFrame("Text", "UIElements.rep.title", UIElements.rep.header)
	UIElements.rep.title:SetText("Debuff Summary")
	UIElements.rep.title:SetLayer(2)
	UIElements.rep.title:SetFontSize(18)
	UIElements.rep.header:SetHeight(y18)
	UIElements.rep.title:SetPoint("TOPLEFT", UIElements.rep.header, "TOPLEFT")

	UIElements.rep.close = UI.CreateFrame("Texture", "UIElements.rep.close", UIElements.rep.header)
	UIElements.rep.close:SetTexture(addon.identifier, "textures/cancel.png")
	UIElements.rep.close:SetPoint("TOPRIGHT", UIElements.rep.header, "TOPRIGHT", 0,0)
	UIElements.rep.close:SetHeight(y18)
	UIElements.rep.close:SetWidth(y18)
	UIElements.rep.close:SetLayer(2)
	UIElements.rep.close:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		_SUMMARYSELECTION = false
		UIElements.rep.header:SetVisible(_SUMMARYSELECTION)
	end, "Event.UI.Input.Mouse.Left.Click")

	UIElements.rep.bkgrnd = UI.CreateFrame("Frame", "UIElements.rep.bkgrnd", UIElements.rep.header)
	UIElements.rep.bkgrnd:SetPoint("TOPLEFT", UIElements.rep.header, "BOTTOMLEFT", 0,0)
	UIElements.rep.bkgrnd:SetBackgroundColor(0.2, 0.2, 0.2, 0.75)
	UIElements.rep.bkgrnd:SetWidth(sw)
	UIElements.rep.bkgrnd:SetHeight((y16*16)+4)
	UIElements.rep.bkgrnd:SetLayer(1)

	UIElements.rep.target = UI.CreateFrame("Text", "UIElements.rep.target", UIElements.rep.header)
	UIElements.rep.target:SetPoint("TOPLEFT", UIElements.rep.bkgrnd, "TOPLEFT", 0,0)
	UIElements.rep.target:SetLayer(2)
	UIElements.rep.target:SetFontSize(16)

	UIElements.rep.time = UI.CreateFrame("Text", "UIElements.rep.time", UIElements.rep.header)
	UIElements.rep.time:SetPoint("TOPLEFT", UIElements.rep.target, "BOTTOMLEFT", 0,0)
	UIElements.rep.time:SetLayer(2)
	UIElements.rep.time:SetFontSize(16)

	local abW = 293-y12

	for k,v in ipairs(_checklist) do
		WATCHDOGWAIT("CreateUI:17:"..k)
		local f = {}
		f.frame = UI.CreateFrame("Frame", "f.frame", UIElements.rep.header)
		f.frame:SetLayer(2)
		f.frame:SetWidth((sw/2)-3)
	
		f.headerbg = UI.CreateFrame("Frame", "f.headerbg", UIElements.rep.header)
		f.headerbg:SetLayer(3)
		f.headerbg:SetWidth((sw/2)-3)
		f.headerbg:SetBackgroundColor(0,0,0.25)
		f.headerbg:SetHeight(y16)
		f.headerbg:SetPoint("TOPLEFT", f.frame, "TOPLEFT", 0,0)
	
		f.header = UI.CreateFrame("Text", "f.header", UIElements.rep.header)
		f.header:SetLayer(4)
		f.header:SetFontSize(16)
		f.header:SetPoint("CENTER", f.headerbg, "CENTER", 0,0)
		f.header:SetText(v.name)
	
		f.uptime = UI.CreateFrame("Text", "f.uptime", UIElements.rep.header)
		f.uptime:SetLayer(4)
		f.uptime:SetFontSize(14)
		f.uptime:SetPoint("TOPLEFT", f.headerbg, "BOTTOMLEFT", 0,0)
		f.uptime:SetText("")
	
		local cnt=0
		local prev=""
	
		for k1, v1 in ipairs(v.list) do
			cnt = cnt+1
			WATCHDOGWAIT("CreateUI:17a:"..k1)
			local d = {}
			d.icon = UI.CreateFrame("Texture", "d.icon", UIElements.rep.header)
			d.icon:SetHeight(y12)
			d.icon:SetWidth(y12)
			d.icon:SetTexture("Rift", _abilityDetails[v1.name].i)
			d.icon:SetLayer(3)
		
			if k1 == 1 then
				d.icon:SetPoint("TOPLEFT", f.uptime, "BOTTOMLEFT", 2,0)
			else
				d.icon:SetPoint("TOPLEFT", prev, "BOTTOMLEFT", 0,0)
			end
			prev = d.icon
		
			d.text = UI.CreateFrame("Text", "d.text", UIElements.rep.header)
			d.text:SetFontSize(12)
			d.text:SetLayer(5)
			d.text:SetPoint("TOPLEFT", d.icon, "TOPRIGHT", 0,0)
		
			d.barbg = UI.CreateFrame("Texture", "d.barbg", UIElements.rep.header)
			d.barbg:SetHeight(y12)
			d.barbg:SetPoint("TOPLEFT", d.text, "TOPLEFT", 0,0)
			d.barbg:SetWidth(abW)
			d.barbg:SetLayer(3)
			d.barbg:SetBackgroundColor(0.5,0,0,0.5)
		
			UIElements.rep.abw = abW

			d.prog = UI.CreateFrame("Texture", "d.barbg", UIElements.rep.header)
			d.prog:SetHeight(y12)
			d.prog:SetPoint("TOPLEFT", d.text, "TOPLEFT", 0,0)
			d.prog:SetLayer(4)
			d.prog:SetBackgroundColor(0,0.5,0,1)
		
			UIElements.rep.abilities[v1.name] = d
		end
	
		f.frame:SetHeight(y16+y14+(cnt*y12))
	
		if k == 1 then
			f.frame:SetPoint("TOPLEFT", UIElements.rep.bkgrnd, "TOPLEFT", 2, (y16*2)+4)
		elseif k == 11 then
			f.frame:SetPoint("TOPLEFT", UIElements.rep.headers[1].frame, "TOPRIGHT", 2,0)
		else
			f.frame:SetPoint("TOPLEFT", UIElements.rep.headers[k-1].frame, "BOTTOMLEFT", 0,2)
		end
	
		UIElements.rep.headers[k] = f
	end
	
	local NUMCAT = 9
	local NUMELE = 23
	
	UIElements.rep.bkgrnd:SetHeight( (NUMCAT*(y16+y14+2)) + (y16*2) + (y12*NUMELE) +4)
end

function RDM.Command_Slash_Register(h, args)
	local r = {}
	local numargs = 0
	for token in string.gmatch(args, "[^%s]+") do
		r[numargs] = token
		numargs=numargs+1
	end
	if numargs > 0 then
		if r[0] == "showdetail" then
			RDMonitor_Settings.showdetail = not RDMonitor_Settings.showdetail
			Debug("Showing detail info: "..tostring(RDMonitor_Settings.showdetail), SILENT)
			UIElements.realtime.frame:SetVisible(RDMonitor_Settings.showdetail)
		elseif r[0] == "debug" then
			if numargs == 2 then
				local dbglvl = tonumber(r[1]) or default_settings.debuglevel
				Debug("Changing debuglevel to: "..dbglvl, SILENT)
				RDMonitor_Settings.debuglevel = dbglvl
				debuglevel = RDMonitor_Settings.debuglevel
			else
				Debug("Current debuglevel: "..debuglevel, SILENT)
			end
		elseif r[0] == "scale" then
			if numargs == 2 then
				local scale = tonumber(r[1]) or 1
				if scale > 0 then
					Debug("Setting scale to: "..scale, SILENT)
					RDMonitor_Settings.scale = scale
					RDM.Scale(scale)
				else
					Debug("Negative scales make no sense!", SILENT)
				end
			else
				Debug("Need a scale value", SILENT)
			end
		elseif r[0] == "mode" then
			if RDMonitor_Settings.trk_target == "focus" then
				RDMonitor_Settings.trk_target = "player.target"
				UIElements.cp.scopetxt:SetText("T")
			elseif RDMonitor_Settings.trk_target == "player.target" then
				RDMonitor_Settings.trk_target = "player.target.target"
				UIElements.cp.scopetxt:SetText("T.T")
			else 
				RDMonitor_Settings.trk_target = "focus"
				UIElements.cp.scopetxt:SetText("F")
			end
			Debug("Switching to use: "..RDMonitor_Settings.trk_target, SILENT)
--			UIElements.tracker.header:SetText(string.format("DEBUFF TRACKER...[%s]", RDMonitor_Settings.trk_target))
		elseif r[0] == "watchdog" then
			Debug("Showing WD yields:", RELEASE)
			for x=1, yieldcnt do
				Debug(string.format("%d @%s [%.4f] %s", x, _watchdogyield[x].tx, _watchdogyield[x].tr, _watchdogyield[x].m), RELEASE)
			end
		elseif r[0] == "orientation" then
			if RDMonitor_Settings.orient == 'H' then
				RDMonitor_Settings.orient = 'V'
			else
				RDMonitor_Settings.orient = 'H'
			end
			RDM.OrderFrames()
			RDM.ResizeCP(RDMonitor_Settings.trx, RDMonitor_Settings.try)
		elseif r[0] == "inspect" then
			local bl = Inspect.Buff.List("player.target") or {}
			local bf = {}
			for k,v in pairs(Inspect.Buff.Detail("player.target", bl)) do
				table.insert(bf, v.name)
			end
			table.sort(bf)
			for k,v in ipairs(bf) do
				print(string.format("%d : %s", k,v))
			end
		end
	end
end

function RDM.Event_System_Secure_Enter(h)
	RDM.UniqMsg("ENTERING SECURE STATE")
	RAIDSTATE = RS_ALIVE
	COMBATSTATE = CB_ACTIVE
	CURFIGHTDTLS.st = Inspect.Time.Real()
end

function RDM.Event_System_Secure_Leave(h)
	RDM.UniqMsg("LEAVING SECURE STATE")
	COMBATSTATE = CB_INACTIVE
end

function RDM.Event_Combat_Death(h,t)
	if t["target"] == CURFIGHTDTLS.targetID then
		Debug("TARGET IS DEAD", STATUSUPDT)
		RDM.EndFight("TARGETDEAD")
	end
	if t["target"] == _playerID then
		_playerDeathTime = os.time()
		PLAYERSTATE = PS_DEAD
		Debug("PLAYERDEAD", STATUSUPDT)
	end

end

function RDM.Event_Unit_Availability_Full(h,t)
	for k,v in pairs(t) do
		if v == "player" then
			local d = Inspect.Unit.Detail("player")
			_playerID = d.id
			Debug("Player = ".._playerID, TESTING)
			
			for k,v in pairs(_abilityDetails) do
				LANG_LOOKUP[v.name] = k
			end
			Command.Event.Detach(Event.Unit.Availability.Full, nil, nil, nil, addon.identifier)
			break
		end
	end
end

function RDM.Event_Addon_SavedVariables_Load_End(h, a)
	if a == addon.identifier then
		SVLoaded = true
		LibVersionCheck.register(addon.toc.Identifier, addon.toc.Version)		
	end
end

---------------------
-- Register Events
---------------------

Command.Event.Attach(Command.Slash.Register("RDM"), RDM.Command_Slash_Register, "Command.Slash.Register")
Command.Event.Attach(Event.Addon.SavedVariables.Load.End, RDM.Event_Addon_SavedVariables_Load_End, "Event.Addon.SavedVariables.Load.End")
Command.Event.Attach(Event.System.Update.Begin, RDM.Event_System_Update_Begin, "Event.System.Update.Begin")
Command.Event.Attach(Event.System.Secure.Enter, RDM.Event_System_Secure_Enter, "Event.System.Secure.Enter")
Command.Event.Attach(Event.System.Secure.Leave, RDM.Event_System_Secure_Leave, "Event.System.Secure.Leave")
Command.Event.Attach(Event.Combat.Death, RDM.Event_Combat_Death, "Event.Combat.Death")
Command.Event.Attach(Event.Unit.Availability.Full, RDM.Event_Unit_Availability_Full, "Event.Unit.Availability.Full")

print(string.format("v%s loaded.", addon.toc.Version))